import logging

from y.contracts import Contract
from ypricemagic.classes import ERC20
from ypricemagic.utils.events import decode_logs, get_logs_asap
from ypricemagic.utils.multicall import fetch_multicall

logger = logging.getLogger(__name__)


class BalancerV2Vault:
    def __init__(self, vault_address) -> None:
        self.address = vault_address
        self.contract = Contract(self.address)
    
    def get_pool_tokens(self, pool_id: int, block=None):
        a = self.contract.getPoolTokens(pool_id, block_identifier = block)
        logger.debug(a)
        return a

    def list_pools(self, block=None):
        topics = ['0x3c13bc30b8e878c53fd2a36b679409c073afd75950be43d8858768e956fbc20e']
        events = decode_logs(get_logs_asap(self.address, topics, to_block=block))
        return {event['poolId'].hex():event['poolAddress'] for event in events}

    def deepest_pool_for(self, token_address, block=None):
        pools = self.list_pools(block=block)
        pools_info = fetch_multicall(*[[self.contract,'getPoolTokens',poolId] for poolId in pools.keys()], block=block)
        deepest_pool = {'pool': None, 'balance': 0}
        for pool, info in zip(pools.values(),pools_info):
            if str(info) == "((), (), 0)": continue
            ct_tokens = len(info[0])
            pool_balances = {info[0][i]: info[1][i] for i in range(ct_tokens)}
            for token, balance in pool_balances.items():
                try:
                    if (
                        token == token_address and balance > deepest_pool['balance']
                        and ERC20(pool).build_name not in ['ConvergentCurvePool','MetaStablePool']
                        ):
                        deepest_pool = {'pool': pool, 'balance': balance}
                except ValueError as e:
                    if 'source code not verified' in str(e): pass
                    else: raise
        return deepest_pool['pool'], deepest_pool['balance']

    
        




from enum import IntEnum

from brownie import chain


class Network(IntEnum):
    Mainnet = 1
    BinanceSmartChain = 56
    xDai = 100
    Polygon = 137
    Fantom = 250
    Arbitrum = 42161
    Avalanche = 43114

    @staticmethod
    def label(chain_id = None):
        if not chain_id: chain_id = chain.id

        if chain_id == Network.Mainnet:                 return 'ETH'
        elif chain_id == Network.BinanceSmartChain:     return 'BSC'
        elif chain_id == Network.xDai:                  return 'XDAI'
        elif chain_id == Network.Fantom:                return 'FTM'
        elif chain_id == Network.Polygon:               return 'POLY'
        elif chain_id == Network.Arbitrum:              return 'ARRB'
        elif chain_id == Network.Avalanche:             return 'AVAX'
    
    @staticmethod
    def name(chain_id = None):
        if not chain_id: chain_id = chain.id

        if chain_id == Network.Mainnet:                 return 'Mainnet'
        elif chain_id == Network.BinanceSmartChain:     return 'Binance Smart Chain'
        elif chain_id == Network.xDai:                  return 'xDai'
        elif chain_id == Network.Fantom:                return 'Fantom'
        elif chain_id == Network.Polygon:               return 'Polygon'
        elif chain_id == Network.Arbitrum:              return 'Arbitrum'
        elif chain_id == Network.Avalanche:             return 'Avalanche'



import logging
from typing import Union

import brownie
from brownie import convert, web3
from eth_typing.evm import Address, BlockNumber
from eth_utils import encode_hex
from eth_utils import function_signature_to_4byte_selector as fourbyte
from y.contracts import Contract

logger = logging.getLogger(__name__)


def _decimals(
    contract_address: Union[str, Address, brownie.Contract, Contract], 
    block: Union[BlockNumber, int, None] = None, 
    return_None_on_failure: bool = False
    ):

    if type(contract_address) != str:
        contract_address = str(contract_address)

    try:
        try: return raw_call(contract_address, "decimals()", block=block, output='int')
        # the contract might not comply with standards, so we can possibly fetch decimals using the non standard abi as a fallback
        except OverflowError: return Contract(contract_address).decimals(block_identifier=block)
        except ValueError as e: 
            if 'execution reverted' in str(e): return Contract(contract_address).decimals(block_identifier=block)
            else: raise
    # if both methods fail to fetch decimals, determine how to fail and fail fast
    except (AttributeError, ValueError) as e:
        fail_msgs = [
            'execution reverted',
            "has no attribute 'decimals'",
        ]
        if return_None_on_failure and any([msg in str(e) for msg in fail_msgs]): return None
        else: raise


def _symbol(
    contract_address: Union[str, Address, brownie.Contract, Contract],
    block: Union[BlockNumber, int, None] = None
    ):

    if type(contract_address) != str:
        contract_address = str(contract_address)

    data = raw_call(contract_address, "symbol()", block=block)
    return convert.to_string(data) 


def _totalSupply(
    contract_address: Union[str, Address, brownie.Contract, Contract], 
    block: Union[BlockNumber, int, None] = None,
    return_None_on_failure: bool = False # TODO: implement this kwarg
    ):

    if type(contract_address) != str:
        contract_address = str(contract_address)

    try:
        return raw_call(contract_address, "totalSupply()", block=block, output='int')
    except OverflowError:
        return Contract(contract_address).totalSupply(block_identifier=block)


def _totalSupplyReadable(
    contract_address: Union[str, Address, brownie.Contract, Contract], 
    block: Union[BlockNumber, int, None] = None
    ):
    return _totalSupply(contract_address,block) / 10 ** _decimals(contract_address,block)


def _balanceOf(
    call_address: Union[str, Address, brownie.Contract, Contract], 
    input_address: Union[str, Address, brownie.Contract, Contract], 
    block: Union[BlockNumber, int, None] = None
    ):

    if type(call_address) != str:
        call_address = str(call_address)
    
    if type(input_address) != str:
        input_address = str(input_address)

    data = raw_call(call_address, "balanceOf(address)", block=block, inputs=input_address)
    try: return convert.to_int(data)
    except: return Contract(call_address).balanceOf(input_address, block_identifier=block)


def _balanceOfReadable(
    call_address: Union[str, Address, brownie.Contract, Contract], 
    input_address: Union[str, Address, brownie.Contract, Contract], 
    block: Union[BlockNumber, int, None] = None
    ):
    return _balanceOf(call_address, input_address, block=block) / 10 ** _decimals(call_address, block)


def raw_call(
    contract_address: Union[str, Address, brownie.Contract, Contract], 
    method: str, 
    block: Union[BlockNumber, int, None] = None, 
    inputs = None, 
    output: str = None
    ):
    '''
    call a contract with only address and method.
    only works with 1 input, ie `balanceOf(address)`
    '''
    data = {'to': convert.to_address(contract_address),'data': prepare_data(method,inputs)}
    response = web3.eth.call(data,block_identifier=block)
    if output is None: return response
    elif output == 'address': return convert.to_address(f'0x{response.hex()[-40:]}')
    elif output in ['int','uint','uint256']: return convert.to_int(response)


def prepare_data(
    method, 
    inputs = None
    ):
    method = encode_hex(fourbyte(method))

    if inputs is None:
        return method

    elif type(inputs) in [str,int,bytes]:
        return method + prepare_input(inputs)
    
    else:
        raise ValueError(f'Unsupported input type. Supported types are: str, int, bytes, Set[str,int,bytes], List[str,int,bytes], Tuple[str,int,bytes], Dict[Any:[str,int,bytes]]')
    
    # these don't work yet, wip
    '''
    elif type(inputs) in (Set,List,Tuple):
        # this doesn't work right now
        inputs = str(concat([method,*[prepare_input(_) for _ in inputs]]))
        logger.warn(inputs)
        return encode_hex(inputs)
    elif type(inputs) == Dict:
        # this doesn't work right now
        inputs = str(concat([method,*[prepare_input(_) for _ in inputs.values()]]))
        logger.warn(inputs)
        return encode_hex(inputs)
    '''


def prepare_input(input):
    try: return '000000000000000000000000' + convert.to_address(input)[2:]
    except: raise 
